@@ -14,15 +14,15 @@
 """Log formatter that enhances tracebacks with extra information.
 """
 
+import io
 import logging
-try:
-    from StringIO import StringIO
-except ImportError: #pragma NO COVER Python3
-    from io import StringIO
 
 from zope.exceptions.exceptionformatter import print_exception
 
 
+Buffer = io.StringIO if bytes is not str else io.BytesIO  # PY2
+
+
 class Formatter(logging.Formatter):
 
     def formatException(self, ei):
@@ -30,7 +30,7 @@ class Formatter(logging.Formatter):
 
         Uses zope.exceptions.exceptionformatter to generate the traceback.
         """
-        sio = StringIO()
+        sio = Buffer()
         print_exception(ei[0], ei[1], ei[2], file=sio, with_filenames=True)
         s = sio.getvalue()
         if s.endswith("\n"):
