@@ -13,9 +13,18 @@
 ##############################################################################
 """ExceptionFormatter tests.
 """
-import unittest
-import doctest
 import sys
+import unittest
+
+
+try:
+    from urllib.error import HTTPError
+except ImportError:
+    # BBB for Python 2.7
+    from urllib2 import HTTPError
+
+
+IS_PY39_OR_GREATER = sys.version_info >= (3, 9)
 
 
 class TextExceptionFormatterTests(unittest.TestCase):
@@ -53,7 +62,6 @@ class TextExceptionFormatterTests(unittest.TestCase):
         self.assertEqual(fmt.getLimit(), 200)
 
     def test_getLimit_sys_has_limit(self):
-        import sys
         fmt = self._makeOne()
         with _Monkey(sys, tracebacklimit=15):
             self.assertEqual(fmt.getLimit(), 15)
@@ -123,13 +131,16 @@ class TextExceptionFormatterTests(unittest.TestCase):
     def test_formatSupplement_w_warnings(self):
         fmt = self._makeOne()
         supplement = DummySupplement()
-        supplement.warnings = ['Beware the ides of March!',
-                               'You\'re gonna get wasted.',
-                              ]
-        self.assertEqual(fmt.formatSupplement(supplement, tb=None),
-                         ['   - Warning: Beware the ides of March!',
-                          '   - Warning: You\'re gonna get wasted.',
-                         ])
+        supplement.warnings = [
+            'Beware the ides of March!',
+            'You\'re gonna get wasted.',
+        ]
+        self.assertEqual(
+            fmt.formatSupplement(supplement, tb=None),
+            [
+                '   - Warning: Beware the ides of March!',
+                '   - Warning: You\'re gonna get wasted.',
+            ])
 
     def test_formatSupplement_w_getInfo_empty(self):
         fmt = self._makeOne()
@@ -152,6 +163,17 @@ class TextExceptionFormatterTests(unittest.TestCase):
         self.assertEqual(fmt.formatTracebackInfo('XYZZY'),
                          '   - __traceback_info__: XYZZY')
 
+    def test_formatTracebackInfo_unicode(self):
+        __traceback_info__ = u"Have a Snowman: \u2603"
+        fmt = self._makeOne()
+
+        result = fmt.formatTracebackInfo(__traceback_info__)
+        expected = '   - __traceback_info__: Have a Snowman: '
+        # utf-8 encoded on Python 2, unicode on Python 3
+        expected += '\xe2\x98\x83' if bytes is str else u'\u2603'
+        self.assertIsInstance(result, str)
+        self.assertEqual(result, expected)
+
     def test_formatLine_no_tb_no_f(self):
         fmt = self._makeOne()
         self.assertRaises(ValueError, fmt.formatLine, None, None)
@@ -168,24 +190,26 @@ class TextExceptionFormatterTests(unittest.TestCase):
         tb.tb_frame = f = DummyFrame()
         lines = fmt.formatLine(tb).splitlines()
         self.assertEqual(len(lines), 1)
-        self.assertEqual(lines[0],
-                         '  File "%s", line %d, in %s'
-                          % (f.f_code.co_filename,
-                             tb.tb_lineno,
-                             f.f_code.co_name,
-                            ))
+        self.assertEqual(
+            lines[0],
+            '  File "%s", line %d, in %s'
+            % (f.f_code.co_filename,
+               tb.tb_lineno,
+               f.f_code.co_name,)
+        )
 
     def test_formatLine_w_f_bogus_linecache_w_filenames(self):
         fmt = self._makeOne(with_filenames=True)
         f = DummyFrame()
         lines = fmt.formatLine(f=f).splitlines()
         self.assertEqual(len(lines), 1)
-        self.assertEqual(lines[0],
-                         '  File "%s", line %d, in %s'
-                          % (f.f_code.co_filename,
-                             f.f_lineno,
-                             f.f_code.co_name,
-                            ))
+        self.assertEqual(
+            lines[0],
+            '  File "%s", line %d, in %s'
+            % (f.f_code.co_filename,
+               f.f_lineno,
+               f.f_code.co_name,)
+        )
 
     def test_formatLine_w_tb_bogus_linecache_wo_filenames(self):
         fmt = self._makeOne(with_filenames=False)
@@ -194,25 +218,27 @@ class TextExceptionFormatterTests(unittest.TestCase):
         f.f_globals['__name__'] = 'dummy.filename'
         lines = fmt.formatLine(tb).splitlines()
         self.assertEqual(len(lines), 1)
-        self.assertEqual(lines[0],
-                         '  Module dummy.filename, line %d, in %s'
-                          % (tb.tb_lineno,
-                             f.f_code.co_name,
-                            ))
+        self.assertEqual(
+            lines[0],
+            '  Module dummy.filename, line %d, in %s'
+            % (tb.tb_lineno,
+               f.f_code.co_name,)
+        )
 
     def test_formatLine_w_f_real_linecache_w_filenames(self):
-        import sys
         fmt = self._makeOne(with_filenames=True)
-        f = sys._getframe(); lineno = f.f_lineno
+        f = sys._getframe()
+        lineno = f.f_lineno
         result = fmt.formatLine(f=f)
         lines = result.splitlines()
         self.assertEqual(len(lines), 2)
-        self.assertEqual(lines[0],
-                         '  File "%s", line %d, in %s'
-                          % (f.f_code.co_filename,
-                             lineno + 1,
-                             f.f_code.co_name,
-                            ))
+        self.assertEqual(
+            lines[0],
+            '  File "%s", line %d, in %s'
+            % (f.f_code.co_filename,
+               lineno + 1,
+               f.f_code.co_name,)
+        )
         self.assertEqual(lines[1],
                          '    result = fmt.formatLine(f=f)')
 
@@ -257,7 +283,38 @@ class TextExceptionFormatterTests(unittest.TestCase):
         err = ValueError('testing')
         self.assertEqual(fmt.formatExceptionOnly(ValueError, err),
                          ''.join(
-                            traceback.format_exception_only(ValueError, err)))
+                             traceback.format_exception_only(ValueError, err)))
+
+    def test_formatExceptionOnly_httperror(self):
+        # On Python 3.11.0 HTTPError may behave wrongly, giving a KeyError in
+        # tempfile when Python tries to format it.
+        # See https://github.com/python/cpython/issues/90113
+        # or examples in Plone tests, especially doctests:
+        # https://github.com/plone/Products.CMFPlone/issues/3663
+        # We don't want to get an error when we format an error,
+        # so let's compensate in our code.
+        fmt = self._makeOne()
+        err = HTTPError('url', 400, 'oops', [], None)
+        result = fmt.formatExceptionOnly(HTTPError, err).strip()
+        # The output can differ too much per Python version,
+        # but it is just one line when stripped.
+        self.assertIn("400", result)
+        self.assertIn("oops", result)
+        self.assertIn("Error", result)
+        self.assertEqual(len(result.splitlines()), 1)
+
+    def test_formatException_httperror(self):
+        # See test_formatExceptionOnly_httperror.
+        # Here we check that formatException works.
+        fmt = self._makeOne()
+        err = HTTPError('url', 400, 'oops', [], None)
+        result = fmt.formatException(HTTPError, err, None)
+        self.assertEqual(result[0], 'Traceback (most recent call last):\n')
+        last = result[-1]
+        # The output can differ per Python version.
+        self.assertIn("400", last)
+        self.assertIn("oops", last)
+        self.assertIn("Error", last)
 
     def test_formatLastLine(self):
         fmt = self._makeOne()
@@ -272,7 +329,7 @@ class TextExceptionFormatterTests(unittest.TestCase):
         self.assertEqual(lines[0], 'Traceback (most recent call last):\n')
         self.assertEqual(lines[1],
                          ''.join(
-                            traceback.format_exception_only(ValueError, err)))
+                             traceback.format_exception_only(ValueError, err)))
 
     def test_formatException_non_empty_tb_stack(self):
         import traceback
@@ -287,26 +344,31 @@ class TextExceptionFormatterTests(unittest.TestCase):
                                    'in dummy_function\n')
         self.assertEqual(lines[2],
                          ''.join(
-                            traceback.format_exception_only(ValueError, err)))
+                             traceback.format_exception_only(ValueError, err)))
 
     def test_formatException_deep_tb_stack_with_limit(self):
         import traceback
-        fmt = self._makeOne(limit=1)
+        fmt = self._makeOne(limit=4)
         err = ValueError('testing')
-        tb0 = DummyTB()
-        tb0.tb_lineno = 27
-        tb0.tb_frame = DummyFrame()
-        tb = DummyTB()
-        tb.tb_frame = DummyFrame()
-        tb.tb_next = tb0
+        tb = self._makeTBs(10)
         lines = fmt.formatException(ValueError, err, tb)
-        self.assertEqual(len(lines), 3)
-        self.assertEqual(lines[0], 'Traceback (most recent call last):\n')
-        self.assertEqual(lines[1], '  Module dummy/filename.py, line 14, '
-                                   'in dummy_function\n')
-        self.assertEqual(lines[2],
-                         ''.join(
-                            traceback.format_exception_only(ValueError, err)))
+
+        self.assertEqual(len(lines), 7)
+
+        expected = [
+            'Traceback (most recent call last):\n',
+            '  Module dummy/filename.py, line 4345, in dummy_function\n',
+            '  Module dummy/filename.py, line 2287, in dummy_function\n',
+            ('...\n'
+             '6 entries omitted, because limit is 4.\n'
+             'Set sys.tracebacklimit or TextExceptionFormatter.limit to'
+             ' a higher value to see omitted entries\n'
+             '...'),
+            '  Module dummy/filename.py, line 26, in dummy_function\n',
+            '  Module dummy/filename.py, line 14, in dummy_function\n',
+            ''.join(traceback.format_exception_only(ValueError, err))
+            ]
+        self.assertEqual(lines, expected)
 
     def test_formatException_recursion_in_tb_stack(self):
         import traceback
@@ -331,19 +393,31 @@ class TextExceptionFormatterTests(unittest.TestCase):
                                    'in dummy_function\n')
         self.assertEqual(lines[4],
                          ''.join(
-                            traceback.format_exception_only(ValueError, err)))
+                             traceback.format_exception_only(ValueError, err)))
 
     def test_extractStack_wo_frame(self):
-        import sys
-        fmt = self._makeOne(limit=1)
-        f = sys._getframe(); lineno = f.f_lineno
+        fmt = self._makeOne()
+        f = sys._getframe()
+        lineno = f.f_lineno
         lines = fmt.extractStack()
-        self.assertEqual(len(lines), 1)
-        self.assertEqual(lines[0], '  Module '
+        # rather don't assert this here
+        # self.assertEqual(len(lines), 10)
+        self.assertEqual(lines[-1], '  Module '
                          'zope.exceptions.tests.test_exceptionformatter, '
                          'line %d, in test_extractStack_wo_frame\n'
                          '    lines = fmt.extractStack()\n' % (lineno + 1))
 
+    def test_extractStack_wo_frame_w_limit(self):
+        fmt = self._makeOne(limit=2)
+        f = sys._getframe()
+        lineno = f.f_lineno
+        lines = fmt.extractStack()
+        self.assertEqual(len(lines), 3)
+        self.assertEqual(lines[-1], '  Module '
+                         'zope.exceptions.tests.test_exceptionformatter, '
+                         'line %d, in test_extractStack_wo_frame_w_limit\n'
+                         '    lines = fmt.extractStack()\n' % (lineno + 1))
+
     def test_extractStack_w_single_frame(self):
         fmt = self._makeOne()
         f = DummyFrame()
@@ -353,24 +427,74 @@ class TextExceptionFormatterTests(unittest.TestCase):
                                    'in dummy_function\n')
 
     def test_extractStack_w_multiple_frames_and_limit(self):
-        fmt = self._makeOne(limit=1)
-        f0 = DummyFrame()
-        f0.f_lineno = 213
-        f = DummyFrame()
-        f.f_back = f0
+        fmt = self._makeOne(limit=2)
+        f = self._makeFrames(10)
         lines = fmt.extractStack(f)
-        self.assertEqual(len(lines), 1)
-        self.assertEqual(lines[0], '  Module dummy/filename.py, line 137, '
-                                   'in dummy_function\n')
+
+        self.assertEqual(len(lines), 3)
+
+        expected = [
+            '  Module dummy/filename.py, line 17, in dummy_function\n',
+            '...8 entries omitted, because limit is 2...\n',
+            '  Module dummy/filename.py, line 1126, in dummy_function\n',
+            ]
+
+        self.assertEqual(expected, lines)
+
+    def test_extractStack_w_recursive_frames(self):
+        fmt = self._makeOne()
+        f = self._makeFrames(3)
+        f.f_back.f_locals['__exception_formatter__'] = 1
+        lines = fmt.extractStack(f)
+        self.assertEqual(len(lines), 4)
+
+        expected = [
+            '  File "dummy/filename.py", line 17, in dummy_function\n',
+            '  File "dummy/filename.py", line 27, in dummy_function\n',
+            ('(Recursive extractStack() stopped,'
+             ' trying traceback.format_stack)\n'),
+            '  Module dummy/filename.py, line 43, in dummy_function\n',
+        ]
+
+        self.assertEqual(expected, lines)
 
     def test_extractStack_w_recursive_frames_and_limit(self):
-        fmt = self._makeOne(limit=1)
-        f = DummyFrame()
-        f.f_back = f
+        fmt = self._makeOne(limit=2)
+        f = self._makeFrames(3)
+        f.f_back.f_locals['__exception_formatter__'] = 1
         lines = fmt.extractStack(f)
-        self.assertEqual(len(lines), 1)
-        self.assertEqual(lines[0], '  Module dummy/filename.py, line 137, '
-                                   'in dummy_function\n')
+        self.assertEqual(len(lines), 3)
+
+        expected = [
+            '  File "dummy/filename.py", line 17, in dummy_function\n',
+            '...2 entries omitted, because limit is 2...\n',
+            '  Module dummy/filename.py, line 43, in dummy_function\n',
+            ]
+
+        self.assertEqual(expected, lines)
+
+    def _makeTBs(self, count):
+        prev = None
+        for _i in range(count):
+            tb = DummyTB()
+            tb.tb_lineno = 14
+            tb.tb_frame = DummyFrame()
+            if prev is not None:
+                tb.tb_lineno = int(prev.tb_lineno * 1.9)
+                tb.tb_next = prev
+            prev = tb
+        return tb
+
+    def _makeFrames(self, count):
+        prev = None
+        for _i in range(count):
+            f = DummyFrame()
+            f.f_lineno = 17
+            if prev is not None:
+                f.f_lineno = int(prev.f_lineno * 1.6)
+                f.f_back = prev
+            prev = f
+        return f
 
 
 class HTMLExceptionFormatterTests(unittest.TestCase):
@@ -401,7 +525,7 @@ class HTMLExceptionFormatterTests(unittest.TestCase):
     def test_escape_w_markup(self):
         fmt = self._makeOne()
         self.assertEqual(fmt.escape('<span>XXX & YYY<span>'),
-                                    '&lt;span&gt;XXX &amp; YYY&lt;span&gt;')
+                         '&lt;span&gt;XXX &amp; YYY&lt;span&gt;')
 
     def test_getPrefix(self):
         fmt = self._makeOne()
@@ -440,12 +564,13 @@ class HTMLExceptionFormatterTests(unittest.TestCase):
         tb = DummyTB()
         tb.tb_frame = f = DummyFrame()
         result = fmt.formatLine(tb)
-        self.assertEqual(result,
-                         '<li>  File "%s", line %d, in %s</li>'
-                          % (f.f_code.co_filename,
-                             tb.tb_lineno,
-                             f.f_code.co_name,
-                            ))
+        self.assertEqual(
+            result,
+            '<li>  File "%s", line %d, in %s</li>'
+            % (f.f_code.co_filename,
+               tb.tb_lineno,
+               f.f_code.co_name,)
+        )
 
     def test_formatLastLine(self):
         fmt = self._makeOne()
@@ -455,7 +580,6 @@ class HTMLExceptionFormatterTests(unittest.TestCase):
 class Test_format_exception(unittest.TestCase):
 
     def _callFUT(self, as_html=False):
-        import sys
         from zope.exceptions.exceptionformatter import format_exception
         t, v, b = sys.exc_info()
         try:
@@ -548,6 +672,7 @@ class Test_format_exception(unittest.TestCase):
     def test_multiple_levels(self):
         # Ensure many levels are shown in a traceback.
         HOW_MANY = 10
+
         def f(n):
             """Produces a (n + 1)-level traceback."""
             __traceback_info__ = 'level%d' % n
@@ -567,23 +692,33 @@ class Test_format_exception(unittest.TestCase):
             pass
         try:
             raise TypeError(C())
-        except:
+        except TypeError:
             s = self._callFUT(True)
-        self.assertTrue(s.find('&lt;') >= 0, s)
-        self.assertTrue(s.find('&gt;') >= 0, s)
+        self.assertIn('&lt;', s)
+        self.assertIn('&gt;', s)
 
     def test_multiline_exception(self):
         try:
             exec('syntax error\n')
-        except Exception:
+        except SyntaxError:
             s = self._callFUT(False)
         lines = s.splitlines()[-3:]
         self.assertEqual(lines[0], '    syntax error')
-        self.assertTrue(lines[1].endswith('    ^')) #PyPy has a shorter prefix
-        self.assertEqual(lines[2], 'SyntaxError: invalid syntax')
+        self.assertIn('    ^', lines[1])
+        self.assertTrue(lines[2].startswith('SyntaxError: invalid syntax'),
+                        lines[2])
+
+    def test_traceback_info_non_ascii(self):
+        __traceback_info__ = u"Have a Snowman: \u2603"
+        try:
+            raise TypeError()
+        except TypeError:
+            s = self._callFUT(True)
+
+        self.assertIsInstance(s, str)
+        self.assertIn('Have a Snowman', s)
 
     def test_recursion_failure(self):
-        import sys
         from zope.exceptions.exceptionformatter import TextExceptionFormatter
 
         class FormatterException(Exception):
@@ -610,16 +745,46 @@ class Test_format_exception(unittest.TestCase):
         self.assertTrue('FormatterException: Formatter failed'
                         in s.splitlines()[-1])
 
+    def test_format_exception_as_html(self):
+        # Test for format_exception (as_html=True)
+        import re
+        from textwrap import dedent
+
+        from zope.exceptions.exceptionformatter import format_exception
+        try:
+            exec('import')
+        except SyntaxError:
+            result = ''.join(format_exception(*sys.exc_info(), as_html=True))
+        expected = dedent("""\
+            <p>Traceback (most recent call last):</p>
+            <ul>
+            <li>  Module {module}, line ABC, in {fn}<br />
+                exec('import')</li>
+            </ul><p>  File "&lt;string&gt;", line 1<br />
+                import<br />
+                     ^<br />
+            SyntaxError: invalid syntax<br />
+            </p>""").format(
+                module='zope.exceptions.tests.test_exceptionformatter',
+                fn='test_format_exception_as_html',
+        )
+        if IS_PY39_OR_GREATER:  # pragma: no cover
+            # Python 3.9 moves the pointer after the statement instead to the
+            # last character of it:
+            expected = expected.replace('^<br />', ' ^<br />')
+        # HTML formatter uses Windows line endings for some reason.
+        result = result.replace('\r\n', '\n')
+        result = re.sub(r'line \d\d\d,', 'line ABC,', result)
+        self.maxDiff = None
+        self.assertEqual(expected, result)
+
 
 class Test_print_exception(unittest.TestCase):
 
     def _callFUT(self, as_html=False):
-        try:
-            from StringIO import StringIO
-        except ImportError:
-            from io import StringIO
-        buf = StringIO()
-        import sys
+        import io
+        buf = io.StringIO() if bytes is not str else io.BytesIO()
+
         from zope.exceptions.exceptionformatter import print_exception
         t, v, b = sys.exc_info()
         try:
@@ -652,7 +817,6 @@ class Test_print_exception(unittest.TestCase):
 class Test_extract_stack(unittest.TestCase):
 
     def _callFUT(self, as_html=False):
-        import sys
         from zope.exceptions.exceptionformatter import extract_stack
         f = sys.exc_info()[2].tb_frame
         try:
@@ -686,7 +850,7 @@ class Test_extract_stack(unittest.TestCase):
 
     def test_traceback_info_html(self):
         try:
-            __traceback_info__ = "Adam & Eve"
+            __traceback_info__ = u"Adam & Eve"
             raise ExceptionForTesting
         except ExceptionForTesting:
             s = self._callFUT(True)
@@ -695,7 +859,7 @@ class Test_extract_stack(unittest.TestCase):
     def test_traceback_supplement_text(self):
         try:
             __traceback_supplement__ = (TestingTracebackSupplement,
-                                        "You're one in a million")
+                                        u"You're one in a million")
             raise ExceptionForTesting
         except ExceptionForTesting:
             s = self._callFUT(False)
@@ -728,8 +892,16 @@ class Test_extract_stack(unittest.TestCase):
         # The warning
         self.assertTrue(s.find("Repent, for the end is nigh") >= 0, s)
 
+    def test_noinput(self):
+        try:
+            raise ExceptionForTesting
+        except ExceptionForTesting:
+            from zope.exceptions.exceptionformatter import extract_stack
+            s = ''.join(extract_stack())
+            self.assertTrue(s.find('test_noinput') >= 0)
 
-class ExceptionForTesting (Exception):
+
+class ExceptionForTesting(Exception):
     pass
 
 
@@ -738,6 +910,7 @@ class TestingTracebackSupplement(object):
     line = 634
     column = 57
     warnings = ['Repent, for the end is nigh']
+
     def __init__(self, expression):
         self.expression = expression
 
@@ -745,33 +918,51 @@ class TestingTracebackSupplement(object):
 class DummySupplement(object):
     def __init__(self, info=''):
         self._info = info
+
     def getInfo(self):
         return self._info
 
 
 class DummyTB(object):
+    # https://docs.python.org/3/reference/datamodel.html#traceback-objects
+    tb_frame = None
     tb_lineno = 14
     tb_next = None
+    tb_lasti = 1
 
 
 class DummyFrame(object):
     f_lineno = 137
     f_back = None
+
     def __init__(self):
         self.f_locals = {}
         self.f_globals = {}
         self.f_code = DummyCode()
 
+
 class DummyCode(object):
     co_filename = 'dummy/filename.py'
     co_name = 'dummy_function'
 
+    def co_positions(self):
+        # New in Python 3.11.
+        # https://docs.python.org/3/reference/datamodel.html#codeobject.co_positions
+        # Note that this is not called for DummyTB if you have tb_lasti=-1.
+        # The 27 in the return value is chosen to match tb_recurse.tb_lineno=27
+        # in test_formatException_recursion_in_tb_stack in this file.
+        # The rest is random.
+        # Note that this code is only called on Python 3.11+, so we mark it for
+        # the coverage tool.
+        return [(27, 2, 3, 4)]  # pragma: no cover
+
+
 class _Monkey(object):
     # context-manager for replacing module names in the scope of a test.
     def __init__(self, module, **kw):
         self.module = module
-        self.to_restore = dict([(key, getattr(module, key, self))
-                                    for key in kw])
+        self.to_restore = {key: getattr(module, key, self)
+                           for key in kw}
         for key, value in kw.items():
             setattr(module, key, value)
 
@@ -780,36 +971,11 @@ class _Monkey(object):
 
     def __exit__(self, exc_type, exc_val, exc_tb):
         for key, value in self.to_restore.items():
-            if value is not self:
+            if value is not self:  # pragma: no cover
                 setattr(self.module, key, value)
             else:
                 delattr(self.module, key)
 
 
-def doctest_format_exception_as_html():
-    """Test for format_exception (as_html=True)
-
-        >>> from zope.exceptions.exceptionformatter import format_exception
-        >>> try:
-        ...     exec('import 2 + 2')
-        ... except:
-        ...     print(''.join(format_exception(*sys.exc_info(), as_html=True)))
-        <p>Traceback (most recent call last):</p>
-        <ul>
-        <li>  Module zope.exceptions.tests.test_exceptionformatter, line 2, in &lt;module&gt;<br />
-            exec('import 2 + 2')</li>
-        </ul><p>  File "&lt;string&gt;", line 1<br />
-            import 2 + 2<br />
-                   ^<br />
-        SyntaxError: invalid syntax<br />
-        </p>
-
-    """
-
-
 def test_suite():
-    return unittest.TestSuite([
-        unittest.defaultTestLoader.loadTestsFromName(__name__),
-        doctest.DocTestSuite(
-                optionflags=doctest.NORMALIZE_WHITESPACE),
-    ])
+    return unittest.defaultTestLoader.loadTestsFromName(__name__)
